/*******************************************************************************
 * This file is part of OpenNMS(R).
 *
 * Copyright (C) 2023 The OpenNMS Group, Inc.
 * OpenNMS(R) is Copyright (C) 1999-2023 The OpenNMS Group, Inc.
 *
 * OpenNMS(R) is a registered trademark of The OpenNMS Group, Inc.
 *
 * OpenNMS(R) is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published
 * by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * OpenNMS(R) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OpenNMS(R).  If not, see:
 *      http://www.gnu.org/licenses/
 *
 * For more information contact:
 *     OpenNMS(R) Licensing <license@opennms.org>
 *     http://www.opennms.org/
 *     http://www.opennms.com/
 *******************************************************************************/

package org.opennms.netmgt.events.api;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.ZonedDateTime;
import java.time.format.DateTimeParseException;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

public class LegacyDatetimeFormatter implements EventDatetimeFormatter {
    public static final ThreadLocal<DateFormat> FORMATTER_FULL = ThreadLocal.withInitial(() -> {
        int timeFormat = DateFormat.FULL;

        // The DateFormat.FULL format for France/Germany do not include the seconds digit
        // which is necessary to have sub-minute resolution in event times. For these
        // locales, we'll fall back to using DateFormat.LONG.
        final String language = Locale.getDefault().getLanguage();
        if (language.equals(Locale.FRANCE.getLanguage()) || language.equals(Locale.GERMANY.getLanguage())) {
            timeFormat = DateFormat.LONG;
        }

        final DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.FULL, timeFormat);
        formatter.setLenient(true);
        return formatter;
    });

    public static final ThreadLocal<DateFormat> FORMATTER_LONG = ThreadLocal.withInitial(() -> {
        final DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.LONG);
        formatter.setLenient(true);
        return formatter;
    });

    public static final ThreadLocal<DateFormat> FORMATTER_FULL_GMT = ThreadLocal.withInitial(() -> {
        int timeFormat = DateFormat.FULL;
        // The DateFormat.FULL format for France/Germany do not include the seconds digit
        // which is necessary to have sub-minute resolution in event times. For these
        // locales, we'll fall back to using DateFormat.LONG.
        final String language = Locale.getDefault().getLanguage();
        if (language.equals(Locale.FRANCE.getLanguage()) || language.equals(Locale.GERMANY.getLanguage())) {
            timeFormat = DateFormat.LONG;
        }

        final DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.FULL, timeFormat);
        formatter.setLenient(true);
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
        return formatter;
    });

    public static final ThreadLocal<DateFormat> FORMATTER_LONG_GMT = ThreadLocal.withInitial(() -> {
        final DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.LONG);
        formatter.setLenient(true);
        formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
        return formatter;
    });

    /**
     * This {@link DateFormat} is used to parse timestamps from XML events that are generated by
     * the send-event.pl script. It always formats timestamps in English so we hard-code the locale
     * as {@link Locale#ENGLISH} in this {@link DateFormat}.
     */
    public static final ThreadLocal<DateFormat> FORMATTER_CUSTOM = ThreadLocal.withInitial(() -> {
        final DateFormat formatter = new SimpleDateFormat("EEEEE, d MMMMM yyyy k:mm:ss 'o''clock' z", Locale.ENGLISH);
        formatter.setLenient(true);
        return formatter;
    });

    public static final ThreadLocal<DateFormat> FORMATTER_DEFAULT = ThreadLocal.withInitial(() -> {
        final DateFormat formatter = DateFormat.getDateTimeInstance();
        formatter.setLenient(true);
        return formatter;
    });

    @Override
    public Date parse(final String dateString) throws ParseException {
        if (dateString == null) {
            throw new ParseException("time was null!", -1);
        }
        try {
            return FORMATTER_LONG.get().parse(dateString);
        } catch (final ParseException parseException) {
            try {
                return FORMATTER_CUSTOM.get().parse(dateString);
            } catch (final ParseException pe) {
                try {
                    return FORMATTER_FULL.get().parse(dateString);
                } catch (final Exception fpe) {
                    try {
                        return Date.from(ZonedDateTime.parse(dateString).toInstant());
                    } catch (final DateTimeParseException dtpe) {
                        throw new ParseException("failed to parse " + dateString + " as an ISO date; giving up", dtpe.getErrorIndex());
                    } catch (final Exception isoe) {
                        throw new ParseException("failed to parse " + dateString + " as an ISO date; giving up", 0);
                    }
                }
            }
        }
    }

    @Override
    public String format(final Date date) {
        return FORMATTER_LONG_GMT.get().format(date);
    }

}
